import math
import os
import pathlib
from typing import Optional

import cached_conv as cc
import gin
import nn_tilde
import torch

from .attention import Embedding, FeatureEmbedding, Prior


class ScriptedPrior(nn_tilde.Module):

    def __init__(
        self,
        run: Optional[str] = None,
        temporal_ratio: int = 1024,
        from_continuous: bool = False,
        vae_path: Optional[str] = None,
        initial_listen: bool = True,
    ) -> None:
        super().__init__()
        print("streaming mode is set to", cc.USE_BUFFER_CONV)

        if run is not None:
            config = os.path.join(run, "config.gin")

            # PARSE CONFIGURATION FILES
            gin.clear_config()
            gin.parse_config_file(config)

        # BUILD MODEL
        model = Prior()

        if run is not None:
            ckpts = pathlib.Path(run).rglob("*.ckpt")
            ckpts = map(str, ckpts)
            ckpts = sorted(ckpts, key=lambda x: "last" in x, reverse=True)
            ckpt = next(iter(ckpts))
            ckpt = torch.load(ckpt, map_location="cpu")["state_dict"]
            model.load_state_dict(ckpt)

        model.eval()

        self.encoder = model.encoder
        self.decoder = model.decoder

        # RETRIEVE INPUT/OUTPUT DETAILS
        self.seq_len = gin.get_bindings(
            "attention.MultiHeadAlibiAttention")["max_seq_len"]
        self.num_rave_quantizers = gin.get_bindings(
            "attention.MultivariateEmbedding")["num_quantizers"]
        self.num_tokens = gin.get_bindings(
            "attention.MultivariateEmbedding")["num_tokens"]
        self.has_encoder = self.encoder is not None

        num_inputs = self.num_rave_quantizers
        self.encoder_input_type = "none"
        self.encoder_num_tokens = 0
        self.feature_vae = None
        self.from_continuous = from_continuous

        if self.has_encoder:
            encoder_embedder = gin.get_bindings(
                "attention.Encoder")["embedder"]
            if isinstance(encoder_embedder(), Embedding):
                num_inputs += 1
                self.encoder_num_tokens = gin.get_bindings(
                    "encoder/attention.Embedding")["num_embeddings"]
                self.encoder_input_type = "discrete"
            elif isinstance(encoder_embedder(), FeatureEmbedding):
                if vae_path is not None:
                    self.encoder_input_type = "vae"
                    self.feature_vae = torch.jit.load(vae_path).eval()
                    num_inputs += self.feature_vae.latent_size
                else:
                    num_inputs += 163  # TODO: put that somewhere in configuration files
                    self.encoder_input_type = "full"
            else:
                raise ValueError(
                    f"Unknown encoder embedder {encoder_embedder}")

            self.encoder_ratio = gin.get_bindings(
                "decoder/attention.TransformerLayer")["encoder_out_ratio"]

        input_labels = [
            f"rave latent {i+1}" for i in range(self.num_rave_quantizers)
        ]

        if self.encoder_input_type == "none":
            pass
        elif self.encoder_input_type == "discrete":
            input_labels.append("semantic tokens")
        elif self.encoder_input_type == "vae":
            input_labels.extend([f"semantic latent {i}" for i in range(8)])
        elif self.encoder_input_type == "full":
            input_labels.extend([f"semantic feature {i}" for i in range(163)])
        else:
            raise ValueError(
                f"Encoder input type {self.encoder_input_type} not understood."
            )

        output_labels = [
            f"rave latent {i+1}" for i in range(self.num_rave_quantizers)
        ]
        output_labels.append("perplexity")

        self.register_buffer(
            "previous_step",
            torch.zeros(
                cc.MAX_BATCH_SIZE,
                1,
                self.num_rave_quantizers,
            ),
        )

        self.register_attribute("temperature", 1.)
        self.register_attribute("listen", initial_listen)
        self.register_attribute("reset", True)

        self.register_method(
            "forward",
            num_inputs,
            temporal_ratio,
            self.num_rave_quantizers + 1,
            temporal_ratio,
            input_labels=input_labels,
            output_labels=output_labels,
        )

        if self.has_encoder:
            # FORCE NN~ RATE DETECTION TO USE LARGER BUFFERS
            self.register_method(
                "dummy_method",
                1,
                self.encoder_ratio * temporal_ratio,
                1,
                self.encoder_ratio * temporal_ratio,
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        batch_size = x.shape[0]
        input_rave = x[:, :self.num_rave_quantizers]

        if self.from_continuous:
            input_rave = self.quantize(input_rave)

        input_rave = torch.clamp(
            input_rave,
            min=0,
            max=self.num_tokens,
        ).long()
        input_rave = input_rave.permute(0, 2, 1)

        if self.reset[0]:
            for n, b in self.named_buffers():
                if "_cache_length" in n or "_relative_index" in n or "_state" in n:
                    b.zero_()
            self.set_reset(False)

        if self.has_encoder:
            assert hasattr(self, "encoder_ratio")
            if self.encoder_input_type == "discrete":
                semantic_tokens = x[:, self.num_rave_quantizers:]
                semantic_tokens = semantic_tokens.reshape(
                    batch_size,
                    semantic_tokens.shape[1],
                    -1,
                    self.encoder_ratio,
                )[..., -1]
                semantic_tokens = torch.clamp(
                    semantic_tokens,
                    0,
                    self.encoder_num_tokens,
                ).long()[:, 0]
                encoder_out = self.encoder(semantic_tokens)
            elif self.encoder_input_type == "vae":
                assert self.feature_vae is not None
                latents = x[:, self.num_rave_quantizers:]
                latents = latents.reshape(
                    batch_size,
                    latents.shape[1],
                    -1,
                    self.encoder_ratio,
                )[..., -1]
                features = self.feature_vae(latents).permute(0, 2, 1)
                encoder_out = self.encoder(features)
            elif self.encoder_input_type == "full":
                features = x[:, self.num_rave_quantizers:]
                features = features.reshape(
                    batch_size,
                    features.shape[1],
                    -1,
                    self.encoder_ratio,
                )[..., -1]
                encoder_out = self.encoder(features.permute(0, 2, 1))
            else:
                raise ValueError(
                    f"encoder input type {self.encoder_input_type} not understood"
                )
        else:
            encoder_out = None

        if self.listen[0]:
            _, perp = self.decoder.sample(
                input_rave,
                encoder_out,
                self.temperature[0],
            )
            self.previous_step[:batch_size].copy_(input_rave[:, -1:])

            if self.from_continuous:
                input_rave = self.dequantize(input_rave)

            out = torch.cat([input_rave, perp[..., None]], -1)
            return out.permute(0, 2, 1)
        else:
            sample_list = []
            perp_list = []

            if encoder_out is not None:
                encoder_out = encoder_out.repeat_interleave(
                    self.encoder_ratio, 1)

            for t in range(input_rave.shape[1]):
                if encoder_out is not None:
                    current_encoder_out = encoder_out[:, t:t + 1]
                else:
                    current_encoder_out = None

                sample, perp = self.decoder.sample(
                    self.previous_step[:batch_size],
                    current_encoder_out,
                    self.temperature[0],
                )

                self.previous_step[:batch_size].copy_(sample)
                sample_list.append(sample)
                perp_list.append(perp)

            samples = torch.cat(sample_list, 1)
            perps = torch.stack(perp_list, 1)

            if self.from_continuous:
                samples = self.dequantize(samples)

            out = torch.cat([samples, perps], -1).permute(0, 2, 1)
            return out

    def dummy_method(self, x: torch.Tensor) -> torch.Tensor:
        # FORCE MINIMUM BUFFER SIZE TO SEMANTIC RATE
        return x

    def quantize(self, x):
        x = .5 * (1 + torch.erf(x / math.sqrt(2)))
        x = torch.floor(x * self.num_tokens - 1).long()
        return x

    def dequantize(self, x):
        x = x.float()
        x = (x + torch.rand_like(x)) / self.num_tokens
        x = torch.erfinv(2 * x - 1) * math.sqrt(2)
        return x

    @torch.jit.export
    def get_temperature(self) -> float:
        return self.temperature[0]

    @torch.jit.export
    def set_temperature(self, temperature: float) -> int:
        if temperature < 0:
            return -1
        self.temperature = (temperature, )
        return 0

    @torch.jit.export
    def get_listen(self) -> bool:
        return self.listen[0]

    @torch.jit.export
    def set_listen(self, listen: bool) -> int:
        self.listen = (listen, )
        return 0

    @torch.jit.export
    def get_reset(self) -> bool:
        return self.reset[0]

    @torch.jit.export
    def set_reset(self, reset: bool) -> int:
        self.reset = (reset, )
        return 0



import os
import pathlib
import shutil
from os import path

import lmdb
from absl import app, flags
from tqdm import tqdm

FLAGS = flags.FLAGS

flags.DEFINE_string(
    'db',
    default=None,
    required=True,
    help='database to compact',
)
flags.DEFINE_integer('max_elm', default=None, help='max number of elements.')


def main(argv):
    db = lmdb.open(FLAGS.db, readonly=True)

    target = path.normpath(FLAGS.db) + "_compact"
    os.makedirs(target, exist_ok=False)

    pretrained = next(iter(pathlib.Path(FLAGS.db).rglob("*.ts")))
    pretrained = str(pretrained)
    pretrained_name = path.basename(pretrained)
    shutil.copy(pretrained, path.join(target, pretrained_name))

    target = lmdb.open(
        target,
        map_size=100 * 1024**3,
    )

    num_elm = 0
    with db.begin() as source_txn:
        keys = list(source_txn.cursor().iternext(values=False))
        with target.begin(write=True) as target_txn:
            for k, v in tqdm(
                    source_txn.cursor().iternext(),
                    desc="Compacting",
                    total=len(keys),
            ):
                target_txn.put(k, v)
                num_elm += 1
                if FLAGS.max_elm is not None:
                    if num_elm >= FLAGS.max_elm:
                        break


if __name__ == "__main__":
    app.run(main)



import cached_conv as cc
import pytest
import torch

from msprior import attention


@torch.enable_grad()
def test_causal():
    cc.use_cached_conv(False)
    seq_len = 128

    layer = attention.MultiHeadAlibiAttention(8)

    x = torch.randn(1, seq_len, 64, requires_grad=True)

    y = layer(x, x, x)
    y[:, seq_len // 2, :].mean().backward()
    grad = x.grad.data[0].sum(-1).reshape(-1)

    assert (grad[seq_len // 2 + 1:] == 0).all()
    assert y.shape == x.shape


@torch.no_grad()
@pytest.mark.parametrize("ratio", [1, 2, 4, 8])
def test_inference(ratio):
    cc.use_cached_conv(False)
    no_cache = attention.MultiHeadAlibiAttention(8, ratio=ratio)

    cc.use_cached_conv(True)
    cache = attention.MultiHeadAlibiAttention(8, ratio=ratio)

    seq_len = 128

    q = torch.randn(1, seq_len, 64)
    k = torch.randn(1, seq_len // ratio, 64)
    v = torch.randn(1, seq_len // ratio, 64)

    pred_no_cache = no_cache(q, k, v)

    for i in range(2):
        pred_cache = []
        cache.reset()

        for i in range(seq_len):
            pred_cache.append(
                cache(q[:, i:i + 1], k[:, i // ratio:i // ratio + 1],
                      v[:, i // ratio:i // ratio + 1]))

        pred_cache = torch.cat(pred_cache, 1)

        assert pred_cache.shape == pred_no_cache.shape
        assert torch.allclose(pred_cache, pred_no_cache, rtol=1e-4, atol=1e-4)



import dataclasses
import functools
import importlib
import itertools
import math
import multiprocessing
import os
import pathlib
import shutil
from datetime import timedelta
from functools import partial
from typing import Sequence, Tuple

import GPUtil
import lmdb
import numpy as np
import torch
import torch.nn as nn
from absl import app, flags
from rave.dataset import normalize_signal
from tqdm import tqdm

import msprior.preprocessor
import msprior.utils

torch.set_grad_enabled(False)

FLAGS = flags.FLAGS

flags.DEFINE_string('rave',
                    default=None,
                    required=True,
                    help='Path to the pretrained rave model')
flags.DEFINE_string('out_path',
                    default=None,
                    required=True,
                    help='Output dataset path')
flags.DEFINE_string('audio',
                    default=None,
                    required=True,
                    help='Path to the audio files')
flags.DEFINE_integer('num_secs', default=8, help='Audio chunk size in seconds')
flags.DEFINE_integer('num_tokens', default=1024, help='Rave vocabulary size')
flags.DEFINE_integer('batch_size',
                     default=64,
                     help='Batch size during processing')
flags.DEFINE_multi_string('preprocessor',
                          default=[],
                          help='Additional preprocessors to import')
flags.DEFINE_integer('gpu', default=None, help='GPU to use.')
flags.DEFINE_integer('db_size',
                     default=100,
                     help='Maximum size of the dataset (GB)')
flags.DEFINE_bool('dyndb',
                     default=True,
                     help='Allow the database to grow dynamically')
flags.DEFINE_bool('normalize',
                  default=False,
                  help="Normalize audio before preprocessing.")
flags.DEFINE_integer('silence_threshold',
                     default=None,
                     help="Optional silence threshold (30 - 50).")
flags.DEFINE_integer(
    'resolution',
    default=32,
    help="Discretization resolution to use when using continuous rave models.")


def process_batch(
    batched_audio_sample: Tuple[int, Sequence[Tuple[np.ndarray, str]]],
    target_size: int,
    database: lmdb.Environment,
    batch_size: int,
) -> int:
    batch_id, audio_sample = batched_audio_sample
    audio_samples, paths = zip(*audio_sample)

    if FLAGS.normalize:
        audio_samples = [normalize_signal(a) for a in audio_samples]

    audio_examples = msprior.preprocessor.preprocess_batch(
        audio_samples, paths, target_size)

    for ae_id, (ae, path) in enumerate(zip(audio_examples, paths)):
        with database.begin(write=True) as txn:
            key = f'{batch_id*batch_size + ae_id:016d}'.encode()
            txn.put(key, ae.SerializeToString())

    return batch_id


@dataclasses.dataclass
class RAVEEncoder(msprior.preprocessor.Preprocessor):
    pretrained_path: str = None

    def __post_init__(self):
        super().__post_init__()
        assert self.pretrained_path is not None
        self.model = torch.jit.load(self.pretrained_path)
        self.model.to(self.device).eval()

        self.temporal_ratio = self.model.encode_params[-1].item()
        self.from_continuous = "Discrete" not in self.model.encoder.original_name
        self.resolution = FLAGS.resolution

    def quantize(self, x):
        x = .5 * (1 + torch.erf(x / math.sqrt(2)))
        x = torch.floor(x * (self.resolution - 1))
        return x

    def __call__(self, audio_arrays, audio_paths):
        audio_arrays = np.stack(audio_arrays, 0)
        audio_batch = torch.from_numpy(audio_arrays).float().reshape(
            audio_arrays.shape[0], 1, -1).to(self.device)
        z = self.model.quantize(audio_batch).permute(0, 2, 1)
        if self.from_continuous:
            z = self.encode(z)
        z = z.cpu().numpy()
        return z


def main(argv):
    if FLAGS.gpu is not None:
        device = torch.device(f'cuda:{FLAGS.gpu}')
    elif available := GPUtil.getAvailable(maxMemory=.05, limit=1):
        device = torch.device(f'cuda:{available[0]}')
    else:
        device = torch.device('cpu')

    msprior.preprocessor.set_default_gpu(device)

    # create rave preprocessor
    FLAGS.rave = os.path.abspath(FLAGS.rave)
    rave_preprocessor = RAVEEncoder(
        'rave',
        mode='dynamic',
        dtype=np.int16,
        pretrained_path=FLAGS.rave,
        resample=False,
    )
    sampling_rate = rave_preprocessor.model.sr
    temporal_ratio = rave_preprocessor.temporal_ratio

    msprior.preprocessor.register_preprocessor(rave_preprocessor)
    msprior.preprocessor.set_sampling_rate(sampling_rate)

    # import additional preprocessors
    for preprocessor in FLAGS.preprocessor:
        importlib.import_module(preprocessor)

    # create data reader and writer
    n_signal = sampling_rate * FLAGS.num_secs
    n_signal = 2**int(math.ceil(math.log2(n_signal)))
    print(f'Using chunks of {n_signal/sampling_rate:.2f}s')

    chunk_load = partial(
        msprior.utils.load_audio_chunk,
        n_signal=n_signal,
        sr=sampling_rate,
        silenceremove=FLAGS.silence_threshold,
    )

    env = lmdb.open(
        FLAGS.out_path,
        map_size=FLAGS.db_size * 1024**3,
        readahead=False,
        writemap=not FLAGS.dyndb,
        map_async=not FLAGS.dyndb,
    )

    shutil.copy(
        FLAGS.rave,
        os.path.join(
            FLAGS.out_path,
            os.path.basename(FLAGS.rave),
        ),
    )

    pool = multiprocessing.Pool()

    # search for audio files
    wavs = itertools.chain(*[
        pathlib.Path(FLAGS.audio).rglob(f"*.{ext}")
        for ext in ['opus', 'ogg', 'wav', 'mp3', 'aiff', 'flac', 'm4a']
    ])
    wavs = list(map(str, wavs))
    print(f'Found {len(wavs)} audio files in {FLAGS.audio}')

    # load and batch chunks
    chunks = msprior.utils.flatmap(
        pool,
        chunk_load,
        wavs,
        queue_size=FLAGS.batch_size,
    )
    chunks_batch = msprior.utils.batch(chunks, batch_size=FLAGS.batch_size)
    chunks_batch = enumerate(chunks_batch)

    # process dataset
    processed_samples = map(
        functools.partial(
            process_batch,
            target_size=n_signal // temporal_ratio,
            database=env,
            batch_size=FLAGS.batch_size,
        ),
        chunks_batch,
    )

    with tqdm() as pbar:
        for batch_id in processed_samples:
            batch_id += 1
            n_seconds = n_signal / sampling_rate * batch_id * FLAGS.batch_size
            pbar.set_description(
                f'dataset length: {timedelta(seconds=n_seconds)}')

    pool.close()
    env.sync()


if __name__ == '__main__':
    app.run(main)



import math
from typing import Callable, Dict, Optional, Sequence, Tuple

import cached_conv as cc
import gin
import pytorch_lightning as pl
import torch
import torch.nn as nn

import msprior
from msprior import utils

TensorDict = Dict[str, torch.Tensor]


class MultiHeadAlibiAttention(nn.Module):

    def __init__(self,
                 n_head: int,
                 max_seq_len: int = 256,
                 ratio: int = 1) -> None:
        super().__init__()
        assert n_head >= 8, f'Alibi needs n_head > 8, got {n_head}'

        self._n_head = n_head
        self._ratio = ratio

        self._cached = cc.USE_BUFFER_CONV
        self._max_seq_len = max_seq_len // ratio

        self.register_cached_attention_bias()
        self.register_buffer("_keys_cache", torch.tensor([]))
        self.register_buffer("_values_cache", torch.tensor([]))
        self.register_buffer("_cache_length", torch.tensor(0))
        self.register_buffer("_relative_index", torch.tensor(0))

    def bias_attention(self, attention: torch.Tensor):
        q_len, k_len = attention.shape[-2:]
        bias = self._cached_attention_bias[..., -q_len:, -k_len:]
        return attention + bias[..., -q_len:, -k_len:]

    def register_cached_attention_bias(self):
        bias = torch.arange(self._max_seq_len)[None] * self._ratio
        bias = bias - torch.arange(self._max_seq_len * self._ratio)[:, None]
        bias = bias.float()
        bias.masked_fill_(bias > 0, -float("inf"))
        bias = bias[None]

        m = torch.ones(self._n_head) * 2**(-8 / self._n_head)
        m = torch.cumprod(m, -1)[:, None, None]
        bias = bias * m
        self.register_buffer('_cached_attention_bias', bias)

    @torch.jit.unused
    def init_kv_cache(self, k: torch.Tensor, v: torch.Tensor):
        self._keys_cache = torch.zeros(
            cc.MAX_BATCH_SIZE,
            k.shape[1],
            self._max_seq_len,
            k.shape[-1],
        ).type_as(k)
        self._values_cache = torch.zeros(
            cc.MAX_BATCH_SIZE,
            v.shape[1],
            self._max_seq_len,
            v.shape[-1],
        ).type_as(v)

    def update_cache(self, k: torch.Tensor, v: torch.Tensor):
        if not len(self._keys_cache) or not len(self._values_cache):
            self.init_kv_cache(k, v)

        input_length = k.shape[2]

        if input_length > self._max_seq_len:
            k = k[:k.shape[0], :, -self._max_seq_len:]
            v = v[:, :, -self._max_seq_len:]
            input_length = self._max_seq_len

        self._keys_cache = self._keys_cache.roll(-input_length, dims=2)
        self._keys_cache[:k.shape[0], :, -input_length:] = k

        self._values_cache = self._values_cache.roll(-input_length, dims=2)
        self._values_cache[:k.shape[0], :, -input_length:] = v

        if self._cache_length < self._max_seq_len:
            self._cache_length += input_length

    def split_heads(self, x: torch.Tensor) -> torch.Tensor:
        x = x.reshape(x.shape[0], x.shape[1], self._n_head, -1)
        x = x.permute(0, 2, 1, 3)
        return x

    def gather_heads(self, x: torch.Tensor) -> torch.Tensor:
        x = x.permute(0, 2, 1, 3)
        x = x.reshape(x.shape[0], x.shape[1], -1)
        return x

    def reset(self):
        self._cache_length.zero_()

    def forward(self,
                q: torch.Tensor,
                k: Optional[torch.Tensor] = None,
                v: Optional[torch.Tensor] = None) -> torch.Tensor:
        autoregressive = q.shape[1] == 1

        kv_available = k is not None or v is not None

        if autoregressive and self._cached:
            kv_available = kv_available and not self._relative_index

        q = self.split_heads(q)

        if kv_available:
            assert k is not None
            assert v is not None
            k = self.split_heads(k)
            v = self.split_heads(v)

        if self._cached:
            if kv_available:
                assert k is not None
                assert v is not None
                self.update_cache(k, v)

            k = self._keys_cache[:q.shape[0]]
            v = self._values_cache[:q.shape[0]]

        assert q is not None
        assert k is not None
        assert v is not None

        content_score = torch.einsum('bhtd,bhsd->bhts', q, k)
        attention = content_score / math.sqrt(q.shape[-1])

        # relative positional embedding
        if autoregressive and self._cached:
            bias_idx = -self._ratio + self._relative_index
            bias = self._cached_attention_bias[:, bias_idx, :]
            attention = attention + bias.unsqueeze(1)
        else:
            attention = self.bias_attention(attention)

        if self._cached and self._cache_length:
            attention[..., :-self._cache_length] = -float('inf')

        if self._cached:
            self._relative_index += q.shape[2]
            self._relative_index = self._relative_index % self._ratio

        attention = torch.softmax(attention, -1)

        out = torch.einsum('bhts,bhsd->bhtd', attention, v)
        out = self.gather_heads(out)

        return out


class FeedForward(nn.Module):

    def __init__(self, input_dim: int) -> None:
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(input_dim, 4 * input_dim),
            nn.ReLU(),
            nn.Linear(4 * input_dim, input_dim),
        )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        return self.net(inputs)


class GRULayer(nn.Module):

    def __init__(self, dropout_rate: float, model_dim: int,
                 num_layers: int) -> None:
        super().__init__()
        self.gru = nn.GRU(
            input_size=model_dim,
            hidden_size=model_dim,
            dropout=dropout_rate,
            num_layers=num_layers,
            batch_first=True,
        )
        self.cached = cc.USE_BUFFER_CONV
        self.register_buffer(
            "_state", torch.zeros(num_layers, cc.MAX_BATCH_SIZE, model_dim))

    def forward(self,
                x: torch.Tensor,
                y: Optional[torch.Tensor] = None) -> torch.Tensor:
        if y is not None:
            raise ValueError(
                f"GRU layer only supported with decoder only mode.")
        if self.cached:
            batch_size = x.shape[0]
            x, state = self.gru(x, self._state[:, :batch_size])
            self._state[:, :batch_size].copy_(state)
        else:
            x, state = self.gru(x)
        return x


class TransformerLayer(nn.Module):

    def __init__(self,
                 attention_factory: Callable[[], MultiHeadAlibiAttention],
                 feed_forward_factory: Callable[[], FeedForward],
                 dropout_rate: float,
                 attend_encoder_out: bool,
                 model_dim: int,
                 head_dim: int,
                 num_heads: int,
                 encoder_out_ratio: Optional[int] = None) -> None:
        super().__init__()

        self.attention_norm = nn.LayerNorm(model_dim)
        self.q_linear = nn.Linear(model_dim, head_dim * num_heads)
        self.kv_linear = nn.Linear(model_dim, 2 * head_dim * num_heads)
        self.attention = attention_factory()
        self.post_attention_linear = nn.Linear(head_dim * num_heads, model_dim)
        self.attention_dropout = nn.Dropout(dropout_rate)

        if attend_encoder_out:
            assert encoder_out_ratio is not None
            self.encoder_attention_norm = nn.LayerNorm(model_dim)
            self.encoder_q_linear = nn.Linear(model_dim, head_dim * num_heads)
            self.encoder_kv_linear = nn.Linear(model_dim,
                                               2 * head_dim * num_heads)
            self.encoder_attention = attention_factory(ratio=encoder_out_ratio)
            self.encoder_post_attention_linear = nn.Linear(
                head_dim * num_heads, model_dim)
            self.encoder_attention_dropout = nn.Dropout(dropout_rate)

        self.feed_forward_norm = nn.LayerNorm(model_dim)
        self.feed_forward = feed_forward_factory()
        self.feed_forward_dropout = nn.Dropout(dropout_rate)

        self.attend_encoder_out = attend_encoder_out

    def forward(self,
                x: torch.Tensor,
                y: Optional[torch.Tensor] = None) -> torch.Tensor:

        # SELF ATTENTION
        x_res = x.clone()
        q = self.q_linear(x)

        k, v = self.kv_linear(x).chunk(2, -1)

        x = self.attention(q, k, v)
        x = self.post_attention_linear(x)
        x = self.attention_dropout(x)
        x = x + x_res
        x = self.attention_norm(x)

        # CROSS ATTENTION
        if self.attend_encoder_out:
            assert hasattr(self, "encoder_q_linear")
            assert y is not None

            x_res = x.clone()
            q = self.encoder_q_linear(x)

            k, v = self.encoder_kv_linear(y).chunk(2, -1)

            x = self.encoder_attention(q, k, v)
            x = self.encoder_post_attention_linear(x)
            x = self.encoder_attention_dropout(x)
            x = x + x_res
            x = self.encoder_attention_norm(x)

        # FEED FORWARD
        x_res = x.clone()
        x = self.feed_forward(x)
        x = self.feed_forward_dropout(x)
        x = x + x_res
        x = self.feed_forward_norm(x)

        return x


class MultivariateEmbedding(nn.Module):

    def __init__(self,
                 num_tokens: int,
                 num_features: int,
                 num_quantizers: int,
                 from_pretrained: Optional[str] = None) -> None:
        super().__init__()
        self.from_pretrained = from_pretrained

        if from_pretrained is not None:
            model = torch.jit.load(from_pretrained).eval()
            embeds = []
            for n, p in model.encoder.rvq.named_buffers():
                if n[-14:] == "codebook.embed":
                    embeds.append(p)
            embeds = torch.cat(embeds, 0)

            self.embedder = nn.Embedding(num_quantizers * num_tokens,
                                         embeds.shape[-1])
            self.embedder.weight.data.copy_(embeds)
            self.proj = nn.Linear(embeds.shape[-1], num_features)
        else:
            self.embedder = nn.Embedding(num_quantizers * num_tokens,
                                         num_features)
            self.proj = None

        self.num_quantizers = num_quantizers
        self.num_tokens = num_tokens

    def forward(self, x: torch.Tensor,
                sum_over_quantizers: bool) -> torch.Tensor:
        if sum_over_quantizers:
            x = x + torch.arange(x.shape[-1]).type_as(x) * self.num_tokens

        x = self.embedder(x.long())

        if sum_over_quantizers:
            x = x.sum(-2)

        if self.from_pretrained is not None:
            x = x.detach()
            x = self.proj(x)

        return x


class Embedding(nn.Embedding):

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        input = input.long()

        return nn.functional.embedding(input, self.weight, self.padding_idx,
                                       self.max_norm, self.norm_type,
                                       self.scale_grad_by_freq, self.sparse)


class FeatureEmbedding(nn.Module):

    def __init__(self, in_features: int, out_features: int) -> None:
        super().__init__()
        self.norm = nn.BatchNorm1d(in_features)
        self.net = nn.Linear(in_features, out_features)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.norm(x.permute(0, 2, 1))
        return self.net(x.permute(0, 2, 1))


class IndividualPredictor(nn.Module):

    def __init__(self, dim: int, num_tokens: int, first_quantizer: bool,
                 dropout: float) -> None:
        super().__init__()
        self.predictor = nn.Sequential(nn.GELU(), nn.Linear(dim, num_tokens))
        self.dropout = nn.Dropout(dropout)

        if not first_quantizer:
            self.film = nn.Linear(dim, 2 * dim)
        else:
            self.film = None

    def forward(self,
                context: torch.Tensor,
                previous_embedding: Optional[torch.Tensor] = None):

        if previous_embedding is not None and self.film is not None:
            film = self.film(self.dropout(previous_embedding))
            scale, bias = torch.chunk(film, 2, -1)
            context = context * scale + bias

        logits = self.predictor(context)

        return logits


class MultivariatePredictor(nn.Module):

    def __init__(
        self,
        dim: int,
        num_quantizers: int,
        num_tokens: int,
        dropout_rate: float,
        shared_embedder: MultivariateEmbedding,
    ) -> None:
        super().__init__()
        self.embedder = shared_embedder

        self.predictors = nn.ModuleList([
            IndividualPredictor(dim, num_tokens, idx == 0, dropout_rate)
            for idx in range(num_quantizers)
        ])
        self.num_tokens = num_tokens
        self.num_quantizers = num_quantizers

    def _forward(
        self,
        context: torch.Tensor,
        targets: Optional[torch.Tensor] = None,
        temperature: float = 1.0
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        all_logits = []
        samples = []

        embedded_token = torch.zeros_like(context)
        for quant_id, predictor in enumerate(self.predictors):
            logits = predictor(context, embedded_token)
            all_logits.append(logits)

            if targets is not None:
                current_token = targets[..., quant_id]
            else:
                current_token = utils.sample_from_logits(logits, temperature)

            samples.append(current_token)

            if quant_id != self.num_quantizers - 1:
                embedded_token += self.embedder(
                    current_token.long() + quant_id * self.num_tokens,
                    sum_over_quantizers=False,
                )

        all_logits = torch.stack(all_logits, -2)
        samples = torch.stack(samples, -1)

        dist = torch.log_softmax(all_logits, -1)
        entropy = -(dist * dist.exp()).sum(-1)
        perplexity = entropy.exp().mean(-1)

        return all_logits, samples, perplexity

    def forward(self,
                context: torch.Tensor,
                targets: Optional[torch.Tensor] = None,
                temperature: float = 1.0) -> torch.Tensor:
        return self._forward(context, targets, temperature)[0]

    def sample(self, x: torch.Tensor,
               temperature: float) -> Tuple[torch.Tensor, torch.Tensor]:
        return self._forward(x, temperature=temperature)[1:]


class Encoder(nn.Module):

    def __init__(
        self,
        embedder: Callable[[], nn.Module],
        layer_factory: Callable[[], TransformerLayer],
        num_layers: int,
    ) -> None:
        super().__init__()
        self.embedder = embedder()

        net = []
        for _ in range(num_layers):
            net.append(layer_factory())
        self.net = nn.ModuleList(net)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.embedder(x)

        for layer in self.net:
            x = layer(x)

        return x


class Decoder(nn.Module):

    def __init__(self, embedder: Callable[[], MultivariateEmbedding],
                 layer_factory: Callable[[], TransformerLayer],
                 predictor: Callable[[], MultivariatePredictor],
                 num_layers: int) -> None:
        super().__init__()
        self.embedder = embedder()

        net = []
        for _ in range(num_layers):
            net.append(layer_factory())
        self.net = nn.ModuleList(net)
        self.streaming = cc.USE_BUFFER_CONV

        self.predictor = predictor(shared_embedder=self.embedder)

    def forward(
        self,
        x: torch.Tensor,
        y: Optional[torch.Tensor] = None,
        targets: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        x = self.embedder(x, sum_over_quantizers=True)

        for layer in self.net:
            x = layer(x, y)

        return self.predictor(x, targets)

    def sample(self,
               x: torch.Tensor,
               y: Optional[torch.Tensor] = None,
               temperature: float = 1.) -> Tuple[torch.Tensor, torch.Tensor]:
        x = self.embedder(x, sum_over_quantizers=True)

        for layer in self.net:
            x = layer(x, y)

        return self.predictor.sample(x, temperature=temperature)


@gin.configurable
class Prior(pl.LightningModule):

    def __init__(
        self,
        decoder_factory: Callable[[], Decoder],
        encoder_factory: Optional[Callable[[], Encoder]] = None,
    ) -> None:
        super().__init__()

        self.encoder = None
        if encoder_factory is not None:
            self.encoder = encoder_factory()
        self.decoder = decoder_factory()

    def forward(
        self,
        x: torch.Tensor,
        y: Optional[torch.Tensor] = None,
        targets: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if self.encoder is not None:
            assert y is not None
            encoder_out = self.encoder(y)
        else:
            encoder_out = None

        return self.decoder(x, encoder_out, targets)

    def sample(self,
               x: torch.Tensor,
               encoder_out: Optional[torch.Tensor] = None,
               temperature: float = 1.) -> Tuple[torch.Tensor, torch.Tensor]:
        if self.encoder is not None:
            assert encoder_out is not None

        return self.decoder.sample(
            x,
            encoder_out,
            temperature=temperature,
        )

    def loss(self, inputs: TensorDict) -> torch.Tensor:
        logits = self.forward(
            inputs["decoder_inputs"],
            inputs["encoder_inputs"],
            inputs["decoder_targets"],
        )

        targets_one_hot = nn.functional.one_hot(
            torch.clamp(inputs["decoder_targets"].long(), 0),
            logits.shape[-1],
        ).float()

        logits = torch.log_softmax(logits, -1)

        cross_entropy = -(logits * targets_one_hot).sum(-1)

        return cross_entropy.mean(), logits

    def training_step(self, batch, batch_idx):
        loss, _ = self.loss(batch)

        self.log('cross_entropy', loss)

        return loss

    def validation_step(self, batch, batch_idx):
        loss, logits = self.loss(batch)

        self.log('val_cross_entropy', loss)

        accuracies = self.accuracy(logits, batch["decoder_targets"])
        for topk, acc in accuracies:
            self.log(f'val_acc_top_{topk}', acc)

    @gin.configurable
    def configure_optimizers(
            self, optimizer_cls: Callable[[], torch.optim.Optimizer],
            scheduler_cls: Callable[[],
                                    torch.optim.lr_scheduler._LRScheduler]):
        optimizer = optimizer_cls(self.parameters())
        scheduler = scheduler_cls(optimizer)

        return [optimizer], [{'scheduler': scheduler, 'interval': 'step'}]

    def accuracy(self, prediction: torch.Tensor,
                 target: torch.Tensor) -> Sequence[Tuple[float, float]]:
        prediction = prediction.cpu()
        target = target.cpu()

        top_10 = torch.topk(prediction, 10, -1).indices
        accuracies = (target[..., None] == top_10).long()
        k_values = [1, 3, 5, 10]
        k_accuracy = []
        for k in k_values:
            current = (accuracies[..., :k].sum(-1) != 0).float()
            k_accuracy.append(current.mean())
        return list(zip(k_values, k_accuracy))

    def on_fit_start(self):
        tb = self.logger.experiment
        config = gin.config_str()
        config = config.split('\n')
        config = ['```'] + config + ['```']
        config = '\n'.join(config)
        tb.add_text("config", config)



from random import randint
from typing import Dict, Tuple

import torch

TensorDict = Dict[str, torch.Tensor]


def decoder_only_rave(inputs: TensorDict, seq_len: int,
                      decoder_key: str) -> TensorDict:
    start = randint(0, inputs[decoder_key].shape[0] - seq_len - 1)
    end = start + seq_len

    input_tokens = inputs[decoder_key][start:end].copy()
    target_tokens = inputs[decoder_key][start + 1:end + 1].copy()

    return {
        "encoder_inputs": torch.Tensor([]),
        "decoder_inputs": input_tokens,
        "decoder_targets": target_tokens,
    }


def encoder_decoder_semantic_token(inputs: TensorDict, seq_len: int,
                                   encoder_key: str,
                                   decoder_key: str) -> TensorDict:
    start = randint(0, inputs[decoder_key].shape[0] - seq_len - 1)
    ratio = inputs[decoder_key].shape[0] // inputs[encoder_key].shape[0]
    start = start - (start % ratio)
    end = start + seq_len

    input_tokens = inputs[decoder_key][start:end].copy()
    target_tokens = inputs[decoder_key][start + 1:end + 1].copy()

    semantic_indices = inputs[encoder_key].copy()
    semantic_indices = semantic_indices[start // ratio:end // ratio]

    return {
        "encoder_inputs": semantic_indices,
        "decoder_inputs": input_tokens,
        "decoder_targets": target_tokens,
    }


def encoder_decoder_semantic_features(inputs: TensorDict, seq_len: int,
                                      encoder_key: str,
                                      decoder_key: str) -> TensorDict:
    start = randint(0, inputs[decoder_key].shape[0] - seq_len - 1)
    ratio = inputs[decoder_key].shape[0] // inputs[encoder_key].shape[0]
    start = start - (start % ratio)
    end = start + seq_len

    input_tokens = inputs[decoder_key][start:end].copy()
    target_tokens = inputs[decoder_key][start + 1:end + 1].copy()

    semantic_features = inputs[encoder_key].copy()
    semantic_features = semantic_features[start // ratio:end // ratio]

    return {
        "encoder_inputs": semantic_features,
        "decoder_inputs": input_tokens,
        "decoder_targets": target_tokens,
    }
